import yaml

from yaml_rulz.errors import YAMLHandlerError
from yaml_rulz.list_handler import ListHandler


class YAMLHandlerBase(object):

    role = ""

    def __init__(self, yml_content, separator=":"):
        self.separator = separator
        parsed_yml = self._import_yml(yml_content)
        self.flat_yml = self._get_flat_dict(parsed_yml)
        self.list_handler = ListHandler(self.flat_yml, separator)
        self.scalars = dict(
            [(key, value) for key, value in self.flat_yml.items() if key not in self.list_handler.list_types]
        )

    def _get_flat_dict(self, parsed_yml):
        if isinstance(parsed_yml, dict):
            return dict(self._flatten_items(parsed_yml.items(), parent=""))
        return dict(self._flatten_items([("", parsed_yml)], parent=""))

    def _flatten_items(self, items, parent):
        flattened_items = []
        for key, value in items:
            prefix = parent + str(key)
            if isinstance(value, list):
                flattened_items.append((prefix, []))
                flattened_items.extend(self._flatten_items(list(enumerate(value)), parent=prefix+self.separator))
            elif isinstance(value, dict):
                flattened_items.extend(self._flatten_items(value.items(), parent=prefix+self.separator))
            else:
                flattened_items.append((prefix, value))
        return flattened_items

    def _import_yml(self, yml_content):
        try:
            return yaml.load(yml_content)
        except yaml.YAMLError as exc:
            raise YAMLHandlerError("Error in {0}\n{1}".format(self.role, exc))


class SchemaHandler(YAMLHandlerBase):

    role = "schema"


class ResourceHandler(YAMLHandlerBase):

    role = "resource"



import re


RE_NUMBER = r"\d+"
RE_EOL = r"$"
RE_LIST_TYPE = r"(^|{0})(\d+)({0}|$)"
RE_LIST_ITEM_SPLITTER = r"(^.*{0})(.*$)"


class ListHandler(object):

    def __init__(self, flat_yml, separator):
        self.separator = separator
        self.list_item_regexp = re.escape(separator) + RE_NUMBER
        self.list_type_regexp = RE_LIST_TYPE.format(separator)
        self.list_types = dict(self._filter_list_types(flat_yml))
        self.groups = self._generate_groups()

    def _filter_list_types(self, flat_yml):
        for key, value in flat_yml.items():
            if re.search(self.list_type_regexp, key):
                yield key, value

    def _generate_groups(self):
        flat_yml_copy = dict(self.list_types.items())
        groups = {}
        for key in self.list_types:
            if key in flat_yml_copy:
                parent = self._get_parent_from_key(key)
                similar_items = self._get_items_starting_with(parent, flat_yml_copy)
                flat_yml_copy = self._remove_items_from_dict(similar_items, flat_yml_copy)
                groups[parent] = dict([(key, value) for key, value in similar_items])
        return groups

    def _get_items_starting_with(self, parent, flat_yaml):
        return [(key, value) for key, value in flat_yaml.items()
                if parent == self._get_parent_from_key(key)]

    def _get_parent_from_key(self, key):
        return re.split(RE_LIST_ITEM_SPLITTER.format(self.list_item_regexp), key)[1]

    @staticmethod
    def _remove_items_from_dict(removable, target_dict):
        for item in removable:
            if item[0] in target_dict:
                del target_dict[item[0]]
        return target_dict



#!/usr/bin/env python

from __future__ import print_function
from argparse import ArgumentParser
import json
import sys

from prettytable import PrettyTable

from yaml_rulz.errors import YAMLHandlerError
from yaml_rulz.validator import YAMLValidator


TABLE_HEADER = ["Severity", "Message", "Schema", "Criterion", "Resource", "Value", "Ref"]


def main():
    args = __parse_arguments()
    has_errors, issues = __read_files_and_call_validator(args.schema, args.resource, args.exclusions)
    if args.raw:
        print(json.dumps(issues, indent=2))
    else:
        __print_error_report(issues)
    if has_errors:
        sys.exit(1)


def __print_error_report(issues):
    table = PrettyTable(TABLE_HEADER)
    for column in TABLE_HEADER:
        table.align[column] = "l"
    for issue in issues:
        table.add_row(
            [issue.get(issue_key.lower()) if issue_key in TABLE_HEADER[:-1]
             else "*" if issue.get(issue_key.lower()) else "" for issue_key in TABLE_HEADER]
        )
    table.sortby = "Severity"
    print(table)


def __read_files_and_call_validator(schema_file, resource_file, exclusions_file):
    schema = __read_file(schema_file)
    resource = __read_file(resource_file)
    exclusions = __read_file(exclusions_file) if exclusions_file else None
    try:
        validator = YAMLValidator(schema, resource, exclusions)
    except YAMLHandlerError as exc:
        print(exc)
        sys.exit(1)
    else:
        return validator.get_validation_issues()


def __parse_arguments():
    argument_parser = ArgumentParser(prog="yaml_rulz")
    argument_parser.add_argument("schema", help="YAML schema file")
    argument_parser.add_argument("resource", help="YAML resource file to be validated")
    argument_parser.add_argument("-x", "--exclusions", help="Exclusions file (optional)")
    argument_parser.add_argument("-r", "--raw", help="Prints the raw error dictionary", action="store_true")
    return argument_parser.parse_args()


def __read_file(filename):
    try:
        with open(filename, "r") as handler:
            return handler.read()
    except (IOError, OSError) as exc:
        print(exc)
        sys.exit(1)


if __name__ == "__main__":
    main()  # pragma: nocover



from unittest import TestCase

from yaml_rulz.rulebook import OmitRule
from yaml_rulz.rulebook import BooleanRule
from yaml_rulz.rulebook import GreaterThanRule
from yaml_rulz.rulebook import LessThanRule
from yaml_rulz.rulebook import RegExpRule
from yaml_rulz.rulebook import PredefinedRegExpRule
from yaml_rulz.rulebook import UniquenessRule


RULE_KEY = "sample:key"
REFERRED_KEY = "other:key"

OMIT_CRITERION = "example"
OMIT_RESOURCE = {
    RULE_KEY: "whatever"
}

BOOLEAN_TRUE_CRITERIA = ["true", "yes", "on"]
BOOLEAN_TRUE_RESOURCE = {
    RULE_KEY: True
}
BOOLEAN_FALSE_CRITERIA = ["false", "no", "off"]
BOOLEAN_FALSE_RESOURCE = {
    RULE_KEY: False
}

GREATER_THAN_CRITERIA_OK = ["99", "0", "90/10+1"]
GREATER_THAN_CRITERIA_FAIL = ["100", "101", "(90/10+1)*20"]
GREATER_THAN_RESOURCE = {
    RULE_KEY: "100"
}

LESS_THAN_CRITERIA_OK = ["101", "924", "1000 + 1000 / 2"]
LESS_THAN_CRITERIA_FAIL = ["100", "-20", "10*4"]
LESS_THAN_RESOURCE = {
    RULE_KEY: "100"
}

REGEXP_RULE_OK = ["^(\\w+)(://)(\\w+)(\\.example\\.com)"]
REGEXP_RULE_FAIL = ["^https://.*"]
REGEXP_RULE_RESOURCE = {
    RULE_KEY: "http://mail.example.com"
}

PREDEF_REGEXP_CRITERIA = ["num", "ipv4", "ipv4_cidr", "ipv6", "ipv6_cidr"]
PREDEF_REGEXP_RESOURCE_OK = {
    "num": [
        {RULE_KEY: "0"},
        {RULE_KEY: "14135"},
        {RULE_KEY: "-12349"},
    ],
    "ipv4": [
        {RULE_KEY: "0.0.0.0"},
        {RULE_KEY: "255.255.255.255"},
        {RULE_KEY: "192.168.0.11"},
    ],
    "ipv4_cidr": [
        {RULE_KEY: "0.0.0.0/0"},
        {RULE_KEY: "255.255.255.255/32"},
        {RULE_KEY: "192.168.0.11/16"},
    ],
    "ipv6": [
        {RULE_KEY: "2001:cdba:0000:0000:0000:0000:3257:9652"},
        {RULE_KEY: "2001:cdba:0:0:0:0:3257:9652"},
        {RULE_KEY: "2001:cdba::3257:9652"},
    ],
    "ipv6_cidr": [
        {RULE_KEY: "2001:cdba:0000:0000:0000:0000:3257:9652/48"},
        {RULE_KEY: "2001:cdba:0:0:0:0:3257:9652/128"},
        {RULE_KEY: "2001:cdba::3257:9652/8"},
    ],
}
PREDEF_REGEXP_RESOURCE_FAIL = {
    "num": [
        {RULE_KEY: "16b"},
        {RULE_KEY: "14.13.5"},
        {RULE_KEY: "-12/349"},
    ],
    "ipv4": [
        {RULE_KEY: "0.0.0.0/3"},
        {RULE_KEY: "259.255.255.255"},
        {RULE_KEY: "192.168.0.341"},
    ],
    "ipv4_cidr": [
        {RULE_KEY: "0.0.0.0/99"},
        {RULE_KEY: "255.255.255.255/32/2"},
        {RULE_KEY: "192.168.0.611/16"},
    ],
    "ipv6": [
        {RULE_KEY: "2001:cdba:0000:0000:0000:0000:3257:9652:9294"},
        {RULE_KEY: "2001:cdba:0:0:0:0:3257:9652/1"},
        {RULE_KEY: "2001:cdba::3257::9652"},
    ],
    "ipv6_cidr": [
        {RULE_KEY: "2001:cdba:0000:0000:0000:0000:3257:9652/192"},
        {RULE_KEY: "2001:cdbg:0:0:0:0:3257:9652/128"},
        {RULE_KEY: "2001:cdba::3257:9652/ff"},
    ],
}

UNIQUENESS_CRITERION = [".*:key"]
UNIQUENESS_RESOURCE_OK = {
    REFERRED_KEY: "100",
    RULE_KEY: "71",
}
UNIQUENESS_RESOURCE_FAIL = {
    REFERRED_KEY: "100",
    RULE_KEY: "100",
}


class TestRulebook(TestCase):

    def test_rule_errors(self):
        rule_errors = [
            (BooleanRule, "tue"),
            (GreaterThanRule, "not number"),
            (LessThanRule, "a124"),
            (PredefinedRegExpRule, "ipv5"),
        ]
        for error in rule_errors:
            rule_class, criterion = error
            rule = rule_class(RULE_KEY, RULE_KEY, criterion)
            result = rule.match(OMIT_RESOURCE)
            self.assertEqual("Error in given criterion", result["message"])

    def test_omit_rule_always_matches(self):
        rule = OmitRule(RULE_KEY, RULE_KEY, OMIT_CRITERION)
        self.assertEqual(None, rule.match(OMIT_RESOURCE))

    def test_boolean_rule_true_positive(self):
        self.__assert_rule(BooleanRule, BOOLEAN_TRUE_CRITERIA, BOOLEAN_TRUE_RESOURCE, None)

    def test_boolean_rule_true_negative(self):
        self.__assert_rule(BooleanRule, BOOLEAN_TRUE_CRITERIA, BOOLEAN_FALSE_RESOURCE,
                           self.__generate_rule_error_dict)

    def test_boolean_rule_false_positive(self):
        self.__assert_rule(BooleanRule, BOOLEAN_FALSE_CRITERIA, BOOLEAN_FALSE_RESOURCE, None)

    def test_boolean_rule_false_negative(self):
        self.__assert_rule(BooleanRule, BOOLEAN_FALSE_CRITERIA, BOOLEAN_TRUE_RESOURCE,
                           self.__generate_rule_error_dict)

    def test_greater_than_rule_positive(self):
        self.__assert_rule(GreaterThanRule, GREATER_THAN_CRITERIA_OK, GREATER_THAN_RESOURCE, None)

    def test_greater_than_rule_negative(self):
        self.__assert_rule(GreaterThanRule, GREATER_THAN_CRITERIA_FAIL, GREATER_THAN_RESOURCE,
                           self.__generate_rule_error_dict)

    def test_less_than_rule_positive(self):
        self.__assert_rule(LessThanRule, LESS_THAN_CRITERIA_OK, LESS_THAN_RESOURCE, None)

    def test_less_than_rule_negative(self):
        self.__assert_rule(LessThanRule, LESS_THAN_CRITERIA_FAIL, LESS_THAN_RESOURCE,
                           self.__generate_rule_error_dict)

    def test_regexp_rule_positive(self):
        self.__assert_rule(RegExpRule, REGEXP_RULE_OK, REGEXP_RULE_RESOURCE, None)

    def test_regexp_rule_negative(self):
        self.__assert_rule(RegExpRule, REGEXP_RULE_FAIL, REGEXP_RULE_RESOURCE,
                           self.__generate_rule_error_dict)

    def test_predef_regexp_rule(self):
        for definition in PREDEF_REGEXP_CRITERIA:
            for resource in PREDEF_REGEXP_RESOURCE_OK.get(definition):
                self.__assert_rule(PredefinedRegExpRule, [definition], resource, None)
            for resource in PREDEF_REGEXP_RESOURCE_FAIL.get(definition):
                self.__assert_rule(PredefinedRegExpRule, [definition], resource,
                                   self.__generate_rule_error_dict)

    def test_uniqueness_positive(self):
        self.__assert_rule(UniquenessRule, UNIQUENESS_CRITERION, UNIQUENESS_RESOURCE_OK, None)

    def test_uniqueness_negative(self):
        self.__assert_rule(UniquenessRule, UNIQUENESS_CRITERION, UNIQUENESS_RESOURCE_FAIL,
                           self.__generate_referred_rule_error_dict)

    def __assert_rule(self, rule_class, criteria, resource, error_generator_callback):
        for criterion in criteria:
            rule = rule_class(RULE_KEY, RULE_KEY, criterion)
            self.assertEqual(None if not error_generator_callback else error_generator_callback(rule, resource),
                             rule.match(resource))

    @staticmethod
    def __generate_rule_error_dict(rule, resource):
        return {
            "schema": RULE_KEY,
            "resource": RULE_KEY,
            "criterion": rule.criterion,
            "value": resource.get(RULE_KEY),
            "message": rule.error_msg,
            "severity": "Error",
            "ref": False,
        }

    @staticmethod
    def __generate_referred_rule_error_dict(rule, resource):
        return {
            "schema": REFERRED_KEY,
            "resource": RULE_KEY,
            "criterion": resource.get(REFERRED_KEY),
            "value": resource.get(RULE_KEY),
            "message": rule.error_msg,
            "severity": "Error",
            "ref": True,
        }



from unittest import TestCase

from yaml_rulz.validator import YAMLValidator


SCHEMA_CONTENT = r"""
---
root:
  key_a: "~ exactly this"
  key_b: "@ num | > 15"
  key_c: "> root:key_b"
"""

RESOURCE_WITH_ISSUES = r"""
---
root:
  key_a: exactly that
  key_b: 6
  key_c: 2
"""

RESOURCE_ALL_OK = r"""
---
root:
  key_a: exactly this
  key_b: 16
  key_c: 20
"""

EXCLUSIONS = r"""
root:key_a
"""

ISSUES = [
    {"resource": "root:key_a",
     "severity": "Warning",
     "value": "exactly that",
     "criterion": "exactly this",
     "schema": "root:key_a",
     "message": "Regular expression mismatch", "ref": False},
    {"criterion": 6,
     "message": "Value must be greater than criterion",
     "ref": True,
     "resource": "root:key_c",
     "severity": "Error",
     "schema": "root:key_b",
     "value": 2},
    {"criterion": "15",
     "message": "Value must be greater than criterion",
     "ref": False,
     "resource": "root:key_b",
     "severity": "Error",
     "schema": "root:key_b",
     "value": 6},
]


class TestValidator(TestCase):

    maxDiff = None

    def test_validator_without_issues(self):
        has_errors, issues = self.__create_validator_and_get_result(RESOURCE_ALL_OK)
        self.assertFalse(has_errors)
        self.assertEqual([], issues)

    def test_validator_with_issues(self):
        has_errors, issues = self.__create_validator_and_get_result(RESOURCE_WITH_ISSUES, EXCLUSIONS)
        self.assertTrue(has_errors)
        try:
            self.assertItemsEqual(ISSUES, issues)
        except AttributeError:
            self.assertEqual(len(ISSUES), len(issues))
            for issue in ISSUES:
                self.assertIn(issue, issues)

    @staticmethod
    def __create_validator_and_get_result(resource, exclusions=None):
        validator = YAMLValidator(SCHEMA_CONTENT, resource, exclusions)
        return validator.get_validation_issues()



try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from unittest import TestCase

try:
    from mock import mock_open
    from mock import patch
except ImportError:
    from unittest.mock import mock_open
    from unittest.mock import patch

from yaml_rulz.cli import main
from yaml_rulz.errors import YAMLHandlerError


DUMMY_FILE_CONTENT = "dummy"
EMPTY_TABLE = """+----------+---------+--------+-----------+----------+-------+-----+
| Severity | Message | Schema | Criterion | Resource | Value | Ref |
+----------+---------+--------+-----------+----------+-------+-----+
+----------+---------+--------+-----------+----------+-------+-----+
"""
EMPTY_RAW = """[]
"""
TABLE_WITH_ISSUE = """+----------+-----------------------------------+---------------------+-----------+---------------\
------+-------+-----+
| Severity | Message                           | Schema              | Criterion | Resource            | Value | Ref |
+----------+-----------------------------------+---------------------+-----------+---------------------+-------+-----+
| Error    | Value must be less than criterion | root:less_than_rule | 1500      | root:less_than_rule | 1500  |     |
+----------+-----------------------------------+---------------------+-----------+---------------------+-------+-----+
"""
RAW_WITH_ISSUE = '[\n  {\n    "resource": "root:less_than_rule", \n    "severity": "Error", \n    "value": 1500, \n' \
                 '    "criterion": "1500", \n    "schema": "root:less_than_rule", \n    "message": "Value must be ' \
                 'less than criterion", \n    "ref": false\n  }\n]\n'
ISSUE = [
    {
        "criterion": "1500",
        "message": "Value must be less than criterion",
        "ref": False,
        "resource": "root:less_than_rule",
        "severity": "Error",
        "schema": "root:less_than_rule",
        "value": 1500,
    },
]


class ArgsNameSpace(object):

    schema = None
    resource = None
    exclusions = None
    raw = None

    def __init__(self, schema, resource, exclusions=None, raw=False):
        self.schema = schema
        self.resource = resource
        self.exclusions = exclusions
        self.raw = raw


class TestCLI(TestCase):
    # pylint: disable = no-member, too-many-instance-attributes

    def setUp(self):
        self.argparser_patcher = patch("yaml_rulz.cli.ArgumentParser")
        self.argparser_mock = self.argparser_patcher.start()
        self.validator_patcher = patch("yaml_rulz.cli.YAMLValidator")
        self.validator_mock = self.validator_patcher.start()
        self.stdout_patcher = patch("sys.stdout", new_callable=StringIO)
        self.stdout_mock = self.stdout_patcher.start()
        self.file_open_patcher = patch("yaml_rulz.cli.open", mock_open(read_data=DUMMY_FILE_CONTENT))
        self.file_open_mock = self.file_open_patcher.start()

    def tearDown(self):
        self.argparser_patcher.stop()
        self.validator_patcher.stop()
        self.stdout_patcher.stop()
        self.file_open_patcher.stop()

    def test_cli_prints_empty_table_when_no_issues(self):
        self.__setup_mocks(ArgsNameSpace("schema", "resource"), (False, []))
        main()
        self.assertEqual(EMPTY_TABLE, self.stdout_mock.getvalue())

    def test_cli_prints_raw_empty_list_when_no_issues(self):
        self.__setup_mocks(ArgsNameSpace("schema", "resource", None, True), (False, []))
        main()
        self.assertEqual(EMPTY_RAW, self.stdout_mock.getvalue())

    def test_cli_prints_issues_in_table(self):
        self.__setup_mocks(ArgsNameSpace("schema", "resource"), (True, ISSUE))
        self.assertRaises(SystemExit, main)
        self.assertEqual(TABLE_WITH_ISSUE, self.stdout_mock.getvalue())

    def test_cli_prints_issues_in_raw_format(self):
        self.__setup_mocks(ArgsNameSpace("schema", "resource", None, True), (True, ISSUE))
        self.assertRaises(SystemExit, main)
        self.assertEqual(
            sorted([line.strip(", ") for line in RAW_WITH_ISSUE.splitlines()]),
            sorted([line.strip(", ") for line in self.stdout_mock.getvalue().splitlines()])
        )

    def test_cli_handles_validator_exceptions(self):
        self.validator_mock.side_effect = YAMLHandlerError
        self.argparser_mock.return_value.parse_args.return_value = ArgsNameSpace("schema", "resource")
        self.assertRaises(SystemExit, main)
        self.assertFalse(self.validator_mock.return_value.get_validation_issues.called)

    def test_cli_handles_file_io_errors(self):
        self.file_open_mock.side_effect = IOError
        self.assertRaises(SystemExit, main)
        self.assertFalse(self.validator_mock.return_value.get_validation_issues.called)

    def __setup_mocks(self, args_ns, val_issues):
        self.argparser_mock.return_value.parse_args.return_value = args_ns
        self.validator_mock.return_value.get_validation_issues.return_value = val_issues



from unittest import TestCase

try:
    from mock import patch
except ImportError:
    from unittest.mock import patch

from yaml_rulz.yaml_handler import YAMLHandlerBase
from yaml_rulz.yaml_handler import YAMLHandlerError


INVALID_YAML = """
---
root:
  key: value
   other_key: other_value
"""
VALID_YAML = """
---
root:
  scalars:
    key_a: value_a
    key_b: value_b
  list:
    - item1
    - item2
  nested_list:
    - id: id1
      inner_list:
        - key1: value1
          key2: value2
        - key1: value1
          key2: value2
    - id: id2
      inner_list:
        - key3: value3
          key4: value4
        - key3: value3
          key4: value4
"""
VALID_YAML_SCALARS = {
    "root:scalars:key_a": "value_a",
    "root:scalars:key_b": "value_b",
    "root:list": [],
    "root:nested_list": [],
}
VALID_YAML_LIST_TYPES = {
    "root:list:0": "item1",
    "root:list:1": "item2",
    "root:nested_list:0:id": "id1",
    "root:nested_list:0:inner_list": [],
    "root:nested_list:0:inner_list:0:key1": "value1",
    "root:nested_list:0:inner_list:0:key2": "value2",
    "root:nested_list:0:inner_list:1:key1": "value1",
    "root:nested_list:0:inner_list:1:key2": "value2",
    "root:nested_list:1:id": "id2",
    "root:nested_list:1:inner_list": [],
    "root:nested_list:1:inner_list:0:key3": "value3",
    "root:nested_list:1:inner_list:0:key4": "value4",
    "root:nested_list:1:inner_list:1:key3": "value3",
    "root:nested_list:1:inner_list:1:key4": "value4",
}


class TestYAMLHandler(TestCase):

    def test_syntax_error_in_yaml_throws_exception(self):
        self.assertRaises(YAMLHandlerError, YAMLHandlerBase, INVALID_YAML)

    @patch("yaml_rulz.yaml_handler.ListHandler")
    def test_handler_with_valid_yaml(self, list_handler_mock):
        list_handler_mock.return_value.list_types = VALID_YAML_LIST_TYPES
        all_data = VALID_YAML_SCALARS.copy()
        all_data.update(VALID_YAML_LIST_TYPES)
        handler = YAMLHandlerBase(VALID_YAML)
        self.assertEqual(VALID_YAML_SCALARS, handler.scalars)
        self.assertEqual(all_data, handler.flat_yml)



# -*- coding: utf-8 -*-

from setuptools import setup


setup(
    name="yaml_rulz",
    version="0.0.1",
    description="A YAML validator",
    license="MIT",
    author="Milan Boleradszki",
    author_email="bmilan1985@gmail.com",
    maintainer="Milan Boleradszki",
    maintainer_email="bmilan1985@gmail.com",
    url="https://github.com/milonoir/yaml_rulz",
    packages=["yaml_rulz"],
    install_requires=["PyYAML", "prettytable"],
    tests_require=["mock"],
    classifiers=[
        # "Development Status :: 1 - Planning",
        # "Development Status :: 2 - Pre-Alpha",
        # "Development Status :: 3 - Alpha",
        "Development Status :: 4 - Beta",
        # "Development Status :: 5 - Production/Stable",
        # "Development Status :: 6 - Mature",
        # "Development Status :: 7 - Inactive",
        "Environment :: Console",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: Unix",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Topic :: Text Processing :: Markup",
    ],
    entry_points={
        "console_scripts": [
            "yaml_rulz = yaml_rulz.cli:main",
        ],
    },
)



import re

from yaml_rulz.errors import RuleError


YAML_TRUE_REGEXP = r"^(true|yes|on)$"
YAML_FALSE_REGEXP = r"^(false|no|off)$"
YAML_KEY_REGEXP = r"^[\w\d:.*+-\\]+$"
NUM_REGEXP = r"-?\d+"
EOL_REGEXP = r"$"
IPV4_REGEXP = r"(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)"
IPV6_REGEXP = r"(" \
              r"([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|" \
              r"([0-9a-fA-F]{1,4}:){1,7}:|" \
              r"([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|" \
              r"([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|" \
              r"([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|" \
              r"([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|" \
              r"([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|" \
              r"[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|" \
              r":((:[0-9a-fA-F]{1,4}){1,7}|:)|" \
              r"fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|" \
              r"::(ffff(:0{1,4}){0,1}:){0,1}" \
              r"((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}" \
              r"(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|" \
              r"([0-9a-fA-F]{1,4}:){1,4}:" \
              r"((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}" \
              r"(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))"
CIDR_32 = r"/(([0-9])|([1-2][0-9])|(3[0-2]))"
CIDR_128 = r"/(([0-9]{1,2})|(1[0-1][0-9])|(12[0-8]))"
ERROR_IN_CRITERION = "Error in given criterion"
ERROR_SEVERITY = "Error"


def raise_rule_error(func):
    def wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception:
            raise RuleError
    return wrapped


def get_rule_response_dict(schema=None, resource=None, criterion=None,  # pylint: disable = too-many-arguments
                           value=None, message=None, ref=None):
    return {
        "schema": schema,
        "resource": resource,
        "criterion": criterion,
        "value": value,
        "message": message,
        "severity": ERROR_SEVERITY,
        "ref": ref,
    }


class RuleBase(object):

    error_msg = ""

    def __init__(self, schema_key, resource_key, criterion):
        self.schema_key = schema_key
        self.resource_key = resource_key
        self.criterion = str(criterion)

    def match(self, resource):
        value = resource.get(self.resource_key)
        self._resolve_references(resource)
        if isinstance(self.criterion, dict):
            if self.resource_key in self.criterion:
                del self.criterion[self.resource_key]
            for location, criterion in self.criterion.items():  # pylint: disable = no-member
                result = self._get_evaluation_result(location, self.resource_key, criterion, value, True)
                if result:
                    return result
        else:
            return self._get_evaluation_result(self.schema_key, self.resource_key, self.criterion, value, False)

    def _get_evaluation_result(self, schema, resource, criterion, value, ref):  # pylint: disable = too-many-arguments
        try:
            if not self._evaluate(criterion, value):
                return get_rule_response_dict(schema=schema,
                                              resource=resource,
                                              criterion=criterion,
                                              value=value,
                                              message=self.error_msg,
                                              ref=ref)
        except RuleError:
            return get_rule_response_dict(schema=schema,
                                          criterion=criterion,
                                          message=ERROR_IN_CRITERION,
                                          ref=ref)

    def _resolve_references(self, resource):
        search_result = []
        for key, value in resource.items():
            if re.match(self.criterion, key):
                search_result.append((key, value))
        self.criterion = dict(search_result) or self.criterion

    def _evaluate(self, criterion, value):
        raise NotImplementedError  # pragma: nocover


class OmitRule(RuleBase):

    def _evaluate(self, criterion, value):
        return True


class BooleanRule(RuleBase):

    error_msg = "Boolean mismatch"

    @staticmethod
    def _eval_criterion(criterion):
        patterns = {
            re.compile(YAML_TRUE_REGEXP, re.IGNORECASE): True,
            re.compile(YAML_FALSE_REGEXP, re.IGNORECASE): False,
        }
        for pattern in patterns:
            if re.match(pattern, criterion):
                return patterns[pattern]
        raise RuleError

    @raise_rule_error
    def _evaluate(self, criterion, value):
        return self._eval_criterion(criterion) == value


class GreaterThanRule(RuleBase):

    error_msg = "Value must be greater than criterion"

    @raise_rule_error
    def _evaluate(self, criterion, value):
        return eval(str(criterion)) < eval(str(value))


class LessThanRule(RuleBase):

    error_msg = "Value must be less than criterion"

    @raise_rule_error
    def _evaluate(self, criterion, value):
        return eval(str(criterion)) > eval(str(value))


class RegExpRule(RuleBase):

    error_msg = "Regular expression mismatch"

    # Regexp-based rules should not try to resolve references
    def _resolve_references(self, resource):
        pass  # pragma: nocover

    @raise_rule_error
    def _evaluate(self, criterion, value):
        return re.match(criterion, str(value)) is not None


class PredefinedRegExpRule(RegExpRule):

    error_msg = "Predefined regular expression mismatch"

    @staticmethod
    def _eval_criterion(criterion):
        patterns = {
            "num": NUM_REGEXP + EOL_REGEXP,
            "ipv4": IPV4_REGEXP + EOL_REGEXP,
            "ipv4_cidr": IPV4_REGEXP + CIDR_32 + EOL_REGEXP,
            "ipv6": IPV6_REGEXP + EOL_REGEXP,
            "ipv6_cidr": IPV6_REGEXP + CIDR_128 + EOL_REGEXP,
        }
        if criterion in patterns:
            return patterns[criterion]
        raise RuleError

    @raise_rule_error
    def _evaluate(self, criterion, value):
        return re.match(self._eval_criterion(criterion), str(value)) is not None


class UniquenessRule(RuleBase):

    error_msg = "Duplicated value"

    @raise_rule_error
    def _evaluate(self, criterion, value):
        return str(criterion) != str(value)



from unittest import TestCase

from yaml_rulz.list_handler import ListHandler


TEST_FLAT_YML_LIST_TYPES = {
    "ericsson:shelf:0:id": "@ num",
    "ericsson:shelf:0:management:address": "@ ipv4",
    "ericsson:shelf:0:management:username": "username",
    "ericsson:shelf:0:management:password": "password",
    "ericsson:shelf:0:blade": [],
    "ericsson:shelf:0:blade:0:id": "@ num",
    "ericsson:shelf:0:blade:0:nic_assignment": "~ whatever",
    "ericsson:shelf:0:blade:1:id": "@ num",
    "ericsson:shelf:0:blade:1:nic_assignment": "~ whatever",
    "ericsson:shelf:0:blade:1:cic:id": "@ num",
    "ericsson:shelf:0:blade:2:id": "@ num",
    "ericsson:shelf:0:blade:2:nic_assignment": "~ whatever",
    "ericsson:shelf:0:blade:2:cic:id": "@ num",
    "ericsson:shelf:0:blade:2:cinder": "",
    "ericsson:shelf:1:id": "@ num",
    "ericsson:shelf:1:blade": [],
    "ericsson:shelf:1:blade:0:id": "@ num",
    "ericsson:shelf:1:blade:0:nic_assignment": "~ whatever",
    "ericsson:simple_list:0": "~ foo",
    "ericsson:simple_list:1": "~ bar",
}
TEST_FLAT_YML_SCALARS = {
    "ericsson:shelf": [],
}
TEST_FLAT_YML = TEST_FLAT_YML_SCALARS.copy()
TEST_FLAT_YML.update(TEST_FLAT_YML_LIST_TYPES)
TEST_GROUPS = {
    "ericsson:shelf:0": {
        "ericsson:shelf:0:id": "@ num",
        "ericsson:shelf:0:blade": [],
        "ericsson:shelf:0:management:address": "@ ipv4",
        "ericsson:shelf:0:management:password": "password",
        "ericsson:shelf:0:management:username": "username",
    },
    "ericsson:shelf:1": {
        "ericsson:shelf:1:id": "@ num",
        "ericsson:shelf:1:blade": [],
    },
    "ericsson:shelf:0:blade:0": {
        "ericsson:shelf:0:blade:0:nic_assignment": "~ whatever",
        "ericsson:shelf:0:blade:0:id": "@ num",
    },
    "ericsson:shelf:0:blade:1": {
        "ericsson:shelf:0:blade:1:nic_assignment": "~ whatever",
        "ericsson:shelf:0:blade:1:id": "@ num",
        "ericsson:shelf:0:blade:1:cic:id": "@ num",
    },
    "ericsson:shelf:0:blade:2": {
        "ericsson:shelf:0:blade:2:nic_assignment": "~ whatever",
        "ericsson:shelf:0:blade:2:cinder": "",
        "ericsson:shelf:0:blade:2:id": "@ num",
        "ericsson:shelf:0:blade:2:cic:id": "@ num",
    },
    "ericsson:shelf:1:blade:0": {
        "ericsson:shelf:1:blade:0:nic_assignment": "~ whatever",
        "ericsson:shelf:1:blade:0:id": "@ num",
    },
    "ericsson:simple_list:0": {
        "ericsson:simple_list:0": "~ foo",
    },
    "ericsson:simple_list:1": {
        "ericsson:simple_list:1": "~ bar",
    },
}


class TestListHandler(TestCase):

    def setUp(self):
        self.handler = ListHandler(TEST_FLAT_YML, ":")

    def test_list_handler_init(self):
        self.assertEqual(TEST_FLAT_YML_LIST_TYPES, self.handler.list_types)
        self.assertEqual(TEST_GROUPS, self.handler.groups)



from itertools import chain
import re

from yaml_rulz.rulebook import get_rule_response_dict
from yaml_rulz.rulebook import BooleanRule
from yaml_rulz.rulebook import GreaterThanRule
from yaml_rulz.rulebook import LessThanRule
from yaml_rulz.rulebook import OmitRule
from yaml_rulz.rulebook import PredefinedRegExpRule
from yaml_rulz.rulebook import RegExpRule
from yaml_rulz.rulebook import UniquenessRule
from yaml_rulz.yaml_handler import ResourceHandler
from yaml_rulz.yaml_handler import SchemaHandler


WARNING_SEVERITY = "Warning"
MISSING_RESOURCE = "Item is missing from resource"
MISSING_SCHEMA = "No rules were found for resource"
MISSING_PROTOTYPE = "No matching prototype was found"
RULE_SEPARATOR_REGEXP = r"\s+\|\s+"
EOL_REGEXP = r"$"


class YAMLValidator(object):

    known_rule_tokens = {
        "*": OmitRule,
        "?": BooleanRule,
        ">": GreaterThanRule,
        "<": LessThanRule,
        "~": RegExpRule,
        "@": PredefinedRegExpRule,
        "!": UniquenessRule,
    }

    def __init__(self, schema_content, resource_content, exclusions_content=None):
        self.schema_handler = SchemaHandler(schema_content)
        self.resource_handler = ResourceHandler(resource_content)
        self.exclusions = YAMLValidator._import_exclusions(exclusions_content)

    def get_validation_issues(self):
        issues = [issue for issue in chain(
            self._find_missing_resource_scalars(),
            self._find_missing_schema_scalars(),
            self._validate_scalars(),
            self._validate_lists(),
        )]
        return self._update_severity(issues)

    def _update_severity(self, issues):
        has_errors = False
        for issue in issues:
            if self._is_excluded(issue["schema"]) or self._is_excluded(issue["resource"]):
                issue["severity"] = WARNING_SEVERITY
            else:
                has_errors = True
        return has_errors, issues

    def _is_excluded(self, key):
        try:
            for exc in self.exclusions:
                if re.match(exc, key):
                    return True
            return False
        except TypeError:
            return False

    def _find_missing_resource_scalars(self):
        for issue in self._yield_missing_scalar_error(self.schema_handler,
                                                      self.resource_handler,
                                                      MISSING_RESOURCE):
            yield issue

    def _find_missing_schema_scalars(self):
        for issue in self._yield_missing_scalar_error(self.resource_handler,
                                                      self.schema_handler,
                                                      MISSING_SCHEMA):
            yield issue

    def _validate_rules(self, flat_schema, flat_resource):
        for resource_key in flat_resource:
            key_mask = self._key_to_mask(self.resource_handler, resource_key)
            for schema_key in flat_schema:
                if re.match(key_mask + EOL_REGEXP, schema_key):
                    rule_chain = self._split_rules(flat_schema[schema_key])
                    for rule_expression in rule_chain:
                        rule = self._get_rule(schema_key, resource_key, rule_expression)
                        # Whole resource must be passed to match() because of possible references in rules
                        result = rule.match(self.resource_handler.flat_yml)
                        if result:
                            yield result

    def _validate_scalars(self):
        for result in self._validate_rules(self.schema_handler.scalars, self.resource_handler.scalars):
            yield result

    def _validate_lists(self):
        for resource_path, resource in self.resource_handler.list_handler.groups.items():
            # Collect prototypes
            path_mask = self._key_to_mask(self.resource_handler, resource_path)
            prototype_candidates = [
                prototype_groups
                for schema_path, prototype_groups in self.schema_handler.list_handler.groups.items()
                if re.match(path_mask + EOL_REGEXP, schema_path)
            ]
            prototypes = self._filter_matching_prototypes(resource, prototype_candidates)
            if not prototypes:
                yield get_rule_response_dict(
                    schema=prototype_candidates,
                    resource=resource_path,
                    message=MISSING_PROTOTYPE,
                )
            # Evaluate prototypes
            prototype_failure_count = 0
            prototype_failures = []
            for prototype in prototypes:
                result = [failure for failure in self._validate_rules(prototype, resource)]
                if result:
                    prototype_failures.extend(result)
                    prototype_failure_count += 1
            if prototype_failure_count >= len(prototypes):
                for failure in prototype_failures:
                    yield failure

    def _filter_matching_prototypes(self, resource, prototypes):
        matching_prototypes = []
        masked_resource_keys = [self._key_to_mask(self.schema_handler, key) for key in resource.keys()]
        for prototype in prototypes:
            masked_prototype_keys = [self._key_to_mask(self.schema_handler, key) for key in prototype.keys()]
            if set(masked_prototype_keys) == set(masked_resource_keys):
                if re.search(self.schema_handler.list_handler.list_item_regexp + EOL_REGEXP, list(prototype)[0]):
                    for key, value in prototype.items():
                        matching_prototypes.append({key: value})
                else:
                    matching_prototypes.append(prototype)
        return matching_prototypes

    def _get_rule(self, schema_key, resource_key, rule_expression):
        try:
            token, criterion = rule_expression.split(" ", 1)
        except (ValueError, AttributeError):
            return OmitRule(schema_key, resource_key, None)
        else:
            if token in self.known_rule_tokens:
                return self.known_rule_tokens[token](schema_key, resource_key, criterion)
            return OmitRule(schema_key, resource_key, None)

    @staticmethod
    def _yield_missing_scalar_error(outer_handler, inner_handler, message):
        for key in outer_handler.scalars:
            if key not in inner_handler.scalars:
                yield get_rule_response_dict(
                    schema=key if outer_handler.role == "schema" else None,
                    resource=key if outer_handler.role == "resource" else None,
                    message=message,
                )

    @staticmethod
    def _import_exclusions(exclusions_content):
        if not exclusions_content:
            return []
        return [exc.strip() for exc in exclusions_content.splitlines() if exc]

    @staticmethod
    def _split_rules(rule_chain):
        return re.split(RULE_SEPARATOR_REGEXP, str(rule_chain))

    @staticmethod
    def _key_to_mask(handler, key):
        return re.sub(handler.list_handler.list_item_regexp, handler.list_handler.list_item_regexp, key)
